// DrMatrix.cpp
// Source for Doctor Matrix member functions

#include <iostream.h>
#include <math.h>
#include <stdlib.h>

#include "drmatrix.hpp"

template <class Type> Vector<Type>::Vector(int n, V_ALIGN vtype=V_HORIZ)
{
	_vtype = vtype;
	_len = n;
	data = (Type*) new Type[n];
	if(data==0)
	{
		cerr << "Not enough memory to allocate vector!" << endl;
		_exit(1);
	}
}

template <class Type> Vector<Type>::Vector(const Vector<Type>& v)
{
	int i;
	_vtype = v._vtype;
	_len = v._len;
	data = (Type*) new Type[v._len];
	if(data==0)
	{
		cerr << "Not enough memory to allocate vector!" << endl;
		_exit(1);
	}
	for(i=0; i<_len; i++)
		data[i] = v.data[i];
}

template <class Type> Vector<Type>& Vector<Type>::operator=(Vector<Type>& v)
{
	int i;
	delete [] data;
	_vtype = v._vtype;
	_len = v._len;
	data = (Type*) new Type[v._len];
	if(data==0)
	{
		cerr << "Not enough memory to allocate vector!" << endl;
		_exit(1);
	}
	for(i=0; i<_len; i++)
		data[i] = v.data[i];
	return (*this);
}

template <class Type> BOOL Vector<Type>::operator==(Vector<Type>& b)
{
	if(_vtype!=b._vtype) return FALSE;
	if(_len!=b._len) return FALSE;
	for(int i=0; i<=_len; i++)
		if((*this)(i) != b(i)) return FALSE;
	return TRUE;
}

template <class Type> Type& Vector<Type>::operator()(int k)
{
	if(k<1 || k>_len)
	{
		cerr << "Subscript out of range - " << k << endl;
		if(k<1)
			return data[0];
		else
			return data[_len-1];
	}
	return data[k-1];
}

template <class Type> Vector<Type> Vector<Type>::T()
{	// transpose
	Vector<Type> v = *this;
	v._vtype = (_vtype==V_HORIZ ? V_VERT : V_HORIZ);
	return v;
}

template <class Type> int Vector<Type>::lead()
{	// returns location of leading (leftmost nonzero) element
	// returns len+1 if vector is the zero vector
	for(int i=0; i<_len; i++)
		if(data[i]!=0)
			return i;
	return _len+1;
}

template <class Type> Type Vector<Type>::abs()
{
	Type res=0;
	for(int i=0; i<_len; i++)
		res+=data[i]*data[i];
	return sqrt(res);
}

template <class Type> Vector<Type> Vector<Type>::operator+(Vector<Type>& b)
{
	Vector<Type> res(_len,_vtype);
	if(_len!=b._len)
	{
		cerr << "Lengths of added vectors must be identical!\n";
		return b;
	}
	if(_vtype!=b._vtype)
	{
		cerr << "Orientation of both vectors in vector sum must be identical!\n";
		return b;
	}
	for(int i=1; i<=_len; i++)
		res(i) = data[i-1]+b(i);
	return res;
}

template <class Type> Vector<Type> Vector<Type>::operator-()
{
	Vector<Type> res(_len,_vtype);
	for(int i=0; i<_len; i++)
		res.data[i] = -data[i];
    return res;
}

template <class Type> Vector<Type> Vector<Type>::operator-(Vector<Type>& b)
{
	Vector<Type> res(_len,_vtype);
	if(_len!=b._len)
	{
		cerr << "Lengths of subtracted vectors must be identical!\n";
		return b;
	}
	if(_vtype!=b._vtype)
	{
		cerr << "Orientation of both vectors in vector subtraction must be identical!\n";
		return b;
	}
	for(int i=1; i<=_len; i++)
		res(i) = data[i-1]-b(i);
	return res;
}

template <class Type> Type Vector<Type>::operator*(Vector<Type>& b)
{
	Type res=0;
	if(_len!=b._len)
	{
		cerr << "Both vectors must be the same length!\n";
		return 0;
	}
	if(_vtype!=V_HORIZ)
	{
		cerr << "Orientation of first vector in vector product must be horizontal!\n";
		return 0;
	}
	if(b._vtype!=V_VERT)
	{
		cerr << "Orientation of second vector in vector product must be vertical!\n";
		return 0;
	}
	for(int i=1; i<=_len; i++)
		res += data[i-1]*b(i);
	return res;
}

template <class Type> ostream& operator<<(ostream& os, Vector<Type>& v)
{
	const char sep = (v._vtype == V_HORIZ ? '\t' : '\n');
	for(int i=1; i<v.len(); i++)
		os << v(i) << sep;
	os << v(v.len());
	return os;
}

template <class Type> istream& operator>>(istream& is,Vector<Type>& v)
{
	for(int i=1; i<=v.len(); i++)
		is >> v(i);
	return is;
}

template <class Type> Vector<Type> operator*(Type a,Vector<Type>& v)
{
	Vector<Type> res(v.len(),v.align());
	for(int i=1; i<=v.len(); i++)
		res(i) = a*v(i);
	return res;
}

template <class Type> Matrix<Type>::Matrix(const Matrix<Type>& m)
{
	int i,j;
	_nrows = m._nrows;
	_ncols = m._ncols;
	data = new Type[_nrows*_ncols];
	if(data==0)
	{
		cerr << "Not enough memory to create matrix!" << endl;
		_exit(1);
	}
	for(i=0; i<_nrows; i++)
		for(j=0; j<_ncols; j++)
			data[i*_ncols+j] = m.data[i*_ncols+j];
}

template <class Type> Matrix<Type>::Matrix(int n,int m)
{
	_nrows = n;
	_ncols = m;
	data = new Type[n*m];
	if(data==0)
	{
		cerr << "Not enough memory to create matrix!" << endl;
		_exit(1);
	}
}

template <class Type> inline Type& Matrix<Type>::operator()(int i,int j)
{
	if(i<1 || i>_nrows || j<1 || j>_ncols)
	{
		cerr << "Subscript out of range - (" << i << "," << j << ")!\n";
		return *data;
	}
	return data[(i-1)*_ncols+j-1];
}

template <class Type> Vector<Type> Matrix<Type>::row(int n)
{
	Vector<Type> v(_ncols,V_HORIZ);
	int en=n;
	if(n<1)
	{
		cerr << "Subscript out of range - Matrix::row(" << n << ")!\n";
		en=1;
	}
	if(n>_nrows)
	{
		cerr << "Subscript out of range - Matrix::row(" << n << ")!\n";
		en=_nrows;
	}
	for(int i=1; i<=_ncols; i++)
		v(i) = (*this)(en,i);
	return v;
}

template <class Type> Vector<Type> Matrix<Type>::col(int m)
{
	Vector<Type> res(_nrows,V_VERT);
	int em=m;
	if(em<1)
	{
		cerr << "Subscript out of range - Matrix::col(" << em << ")!\n";
		em = 1;
	}
	if(em>_ncols)
	{
		cerr << "Subscript out of range - Matrix::col(" << em << ")!\n";
		em = _ncols;
	}
	for(int i=1; i<=_nrows; i++)
		res(i) = (*this)(i,em);
	return res;
}

template <class Type> Matrix<Type> Matrix<Type>::T()
{	// Transpose function
	Matrix<Type> res(_ncols,_nrows); // switched on purpose
	int n,m;
	for(m=1; m<=_ncols; m++)
		for(n=1; n<=_nrows; n++)
			res(m,n) = (*this)(n,m);
	return res;
}

template <class Type> Type Matrix<Type>::tr()
{	// trace function - returns sum of diagonal
	Type res=0;
	if(!square())
	{
		cerr << "Trace function can only be used on square matrices!\n";
		return res;
	}
	for(int i=1; i<=_nrows; i++)
		res += (*this)(i,i);
	return res;
}

template <class Type> BOOL Matrix<Type>::diag()
{	// returns TRUE if matrix is diagonal
	int n,m;
	if(!square()) return FALSE;
	for(n=1; n<=_nrows; n++)
		for(m=1; m<=_ncols; m++)
			if(m!=n)
				if((*this)(n,m)!=0)
					return FALSE;
	return TRUE;
}

template <class Type> BOOL Matrix<Type>::iden()
{	// returns TRUE if matrix is the identity matrix
	if(!diag())
		return FALSE;
	for(int n=1; n<=_nrows; n++)
		if((*this)(n,n)!=1)
			return FALSE;
	return TRUE;
}

template <class Type> BOOL Matrix<Type>::scalar()
{	// returns TRUE if matrix is scalar, i.e. can be written as kI,
	// where k is real and I is the identity matrix of that order
	Type t;
	if(!diag())
		return FALSE;
	t = *data; // equivalent to t=(*this)(1,1);
	for(int i=2; i<=_nrows; i++)
		if((*this)(i,i) != t)
			return FALSE;
	return TRUE;
}

template <class Type> BOOL Matrix<Type>::sym()
{	// returns TRUE if matrix is symmetric, i.e. m(i,j)==m(j,i) for all i,j
	int i,j;
	if(!square())
		return FALSE;
	for(i=1; i<=_nrows; i++)
		for(j=1; j<i; j++)
			if((*this)(i,j) != (*this)(j,i))
				return FALSE;
	return TRUE;
}

template <class Type> BOOL Matrix<Type>::antisym()
{	// returns TRUE is matrix is antisymmetric,
	// i.e.  m(i,j) == -m(j,i)  for all i,j
	int i,j;
	if(!square())
		return FALSE;
	for(i=1; i<=_nrows; i++)
		for(j=1; j<=i; j++)
			if((*this)(i,j) != -((*this)(j,i)))
				return FALSE;
	return TRUE;
}

template <class Type> BOOL Matrix<Type>::square()
{	// returns true if matrix is square, i.e. nrows()==ncols()
	if(_nrows==_ncols)
		return TRUE;
	else
		return FALSE;
}

template <class Type> Matrix<Type>& Matrix<Type>::operator=(Matrix<Type>& b)
{
	delete [] data;
	data = new Type[b._nrows*b._ncols];
	_nrows = b._nrows;
	_ncols = b._ncols;
	for(int i=0; i<_nrows*_ncols; i++)
		data[i] = b.data[i];
	return (*this);
}

template <class Type> Matrix<Type> Matrix<Type>::operator+(Matrix<Type>& b)
{
	Matrix<Type> res(_nrows,_ncols);
	int n,m;
	if(_nrows!=b._nrows || _ncols!=b._ncols)
	{
		cerr << "Added matrices must have identical dimensions!\n";
		return b;
	}
	for(n=1; n<=_nrows; n++)
		for(m=1; m<=_ncols; m++)
			res(n,m) = (*this)(n,m) + b(n,m);
    return res;
}

template <class Type> Matrix<Type> Matrix<Type>::operator-(Matrix<Type>& b)
{
	Matrix<Type> res(_nrows,_ncols);
	int n,m;
	if(_nrows!=b._nrows || _ncols!=b._ncols)
	{
		cerr << "Subtracted matrices must have identical dimensions!\n";
		return b;
	}
	for(n=1; n<=_nrows; n++)
		for(m=1; m<=_ncols; m++)
			res(n,m) = (*this)(n,m) - b(n,m);
    return res;
}

template <class Type> Matrix<Type> Matrix<Type>::operator-()
{
	Matrix<Type> res(_nrows,_ncols);
	int i,j;
	for(i=1; i<=_nrows; i++)
		for(j=1; j<=_ncols; j++)
			res(i,j) = -(*this)(i,j);
	return res;
}

template <class Type> Matrix<Type> Matrix<Type>::operator*(Matrix<Type>& b)
{
	int i,j,n;
	Matrix<Type> res(_nrows,b._ncols);
	if(_ncols != b._nrows)
	{
		cerr << "Undefined matrix multiplication!\n";
		return b;
	}
	for(i=1; i<=res._nrows; i++)
		for(j=1; j<=res._ncols; j++)
		{
			res(i,j)=0;
			for(n=1; n<=_ncols; n++)
				res(i,j) += (*this)(i,n) * b(n,j);
		}
	return res;
}

template <class Type> BOOL Matrix<Type>::operator==(Matrix<Type>& b)
{
	int i,j;
	if(_nrows!=b._nrows) return FALSE;
	if(_ncols!=b._ncols) return FALSE;
	for(i=1; i<=_nrows; i++)
		for(j=1; j<=_ncols; j++)
			if((*this)(i,j) != b(i,j)) return FALSE;
	return TRUE;
}

template <class Type> Matrix<Type> operator*(Type a,Matrix<Type>& b)
{
	int i,j;
	Matrix<Type> res(b.nrows(),b.ncols());
	for(i=1; i<=b.nrows(); i++)
		for(j=1; j<=b.ncols(); j++)
			res(i,j) = a * b(i,j);
	return res;
}

template <class Type> ostream& operator<<(ostream& os, Matrix<Type>& m)
{
	int i,j;
	for(i=1; i<=m.nrows(); i++)
	{
		for(j=1; j<m.ncols(); j++)
			os << m(i,j) << '\t';
		os << m(i,m.ncols()) << '\n';
	}
	return os;
}

template <class Type> istream& operator>>(istream& is, Matrix<Type>& m)
{
	int i,j;
	for(i=1; i<=m.nrows(); i++)
		for(j=1; j<=m.ncols(); j++)
			is >> m(i,j);
	return is;
}

template <class Type> Type Matrix<Type>::det()
{
	int n,i,j;
	Type res;
	if(!square())
	{
		cerr << "Can't calculate determinant of non-square matrix!\n";
		return 0;
	}
	if(_nrows==2)
		return data[0]*data[3] - data[1]*data[2];
	if(_nrows==3) // quicker than using too much recursion
		return data[0]*(data[4]*data[8]-data[5]*data[7]) -
			   data[1]*(data[3]*data[8]-data[5]*data[6]) +
			   data[2]*(data[3]*data[7]-data[4]*data[6]);
	// nrows>3, so there's no choice except recursion
	Matrix<Type> m(_nrows-1,_ncols-1);
	for(n=1,res=0; n<=_nrows; n++)
	{
		for(i=2; i<=_nrows; i++)
		{
			for(j=1; j<n; j++)
				m(i-1,j)=(*this)(i,j);
			for(j=n+1; j<=_ncols; j++)
				m(i-1,j-1)=(*this)(i,j);
		}
		if(n%2==1)
			res += (*this)(1,n) * m.det();
		else
			res -= (*this)(1,n) * m.det();
	}
	return res;
}

template <class Type> int Matrix<Type>::lead(int row)
{	// returns location of leading element, ncols+1 if blank line, -1 on error
	if(row<1 || row>_nrows)
	{
		cerr << "Subscript out of range - Matrix::lead(" << row << ")!\n";
		return -1;
	}
	for(int i=1; i<=_ncols; i++)
		if((*this)(row,i)!=0)
			return i;
	return _ncols+1;
}

template <class Type> BOOL Matrix<Type>::stepped()
{	// returns TRUE if all nonzero rows satisfy lead(i)>lead(j) for all i>j,
	// and zero rows (if any) are found only at the bottom of the matrix
	int n,l2=lead(1),l1;
	for(n=2; n<=_nrows; n++)
	{
		l1 = l2;
		l2 = lead(n);
		if(l2<l1) return FALSE;
		if(l1==l2 && l2!=_ncols+1) return FALSE;
	}
	return TRUE;
}

int _matrix_compare(const void* a,const void* b)
{
	const int a1 = *(int*)a, b1 = *(int*)b;
	if(a1<b1) return -1;
	if(a1>b1) return 1;
	return 0;
}

template <class Type> Matrix<Type> Matrix<Type>::step()
{	// returns an equivalent, stepped matrix
	int i,n,fn,fn1,ln;
	Matrix<Type> m=(*this), mtag=(*this);
	int* array = (int*) new int[m._nrows*2];
	if(array==0)
	{
		cerr << "Not enough memory!\n";
		_exit(1);
	}

	while(!m.stepped()) {
		for(n=1; n<=m._nrows; n++) {
			array[2*n-2] = m.lead(n);
			array[2*n-1] = n;
		}
		qsort(array,m._nrows,2*sizeof(int),_matrix_compare);
		for(n=1; n<=m._nrows; n++)
			for(i=1; i<=m._ncols; i++)
				mtag(n,i) = m(array[2*n-1],i);
		for(n=1; n<=m._nrows; n++)
			for(i=1; i<=m._ncols; i++)
				m(n,i) = mtag(n,i);
		for(n=2; n<=m._nrows; n++)
		{
			ln = m.lead(n);
			if(ln>m.lead(n-1)) continue; // so far, so good!
			if(ln==m.lead(n-1) && ln==m._ncols+1) continue; // also ok
			if(ln<m.lead(n-1)) break; // re-sort required
			fn = m(n-1,ln);
			fn1= -m(n,ln);
			for(i=ln; i<=m._ncols; i++)
				m(n,i) = fn*m(n,i) + fn1*m(n-1,i);
		}
	}
	delete [] array;
	return m;
}

template <class Type> Matrix<Type> Matrix<Type>::canon()
{	// returns the equivalent canonized matrix
	// does not work for integers
	Matrix<Type> m = step(); // First, get an equivalent stepped matrix
	int i,j,k,n;
	Type t;

	// Now, make sure all leading elements are 1
	for(i=1; i<=m._nrows; i++)
	{
		k = m.lead(i);
		if(k>m._ncols) continue;
		t = m(i,k);
		if(t!=1)
			for(j=1; j<=m._ncols; j++)
				m(i,j) /= t;
	}

	// Then, make sure the columns containing leading elements contain
	// only zeroes, except for the leading element itself
	for(i=1; i<=m._nrows; i++)
	{
		k = m.lead(i);
		if(k>m._ncols) continue;
		for(j=1; j<i; j++)
			if(m(j,k)!=0)
			{
				t = m(j,k);
				for(n=1; n<=m._ncols; n++)
					m(j,n) = m(j,n) - t*m(i,n);
			}
	}
	return m;
}

template <class Type> int Matrix<Type>::rank()
{	// return the rank of the matrix
	Matrix<Type> m = step();
	for(int i=m._nrows; i>=1; i--)
		if(m.lead(i)!=_ncols+1)
			return i;
	return 0; // just to keep the compiler happy
}

template <class Type> BOOL Matrix<Type>::reg()
{	// regular (reversible) matrix
	if(!square()) return FALSE;
	if(rank()!=_nrows) return FALSE;
	return TRUE;
}

template <class Type> BOOL Matrix<Type>::sing()
{	// singular (non-reversible) matrix
	if(!square()) return FALSE;
	if(rank()==_nrows) return FALSE;
	return TRUE;
}

template <class Type> Matrix<Type> Matrix<Type>::inv()
{	// returns the inverse matrix, or the zero matrix if matrix is singular
	Matrix<Type> zero(1,1);
	int i,j;
	zero(1,1)=0;
	if(!square())
	{
		cerr << "Attempting to invert a non-square matrix!\n";
		return zero;
	}
	Matrix<Type> M(_nrows,_nrows*2);
	for(i=1; i<=_nrows; i++)
		for(j=1; j<=_nrows; j++)
		{
			M(i,j) = (*this)(i,j);
			M(i,j+_nrows) = (i==j ? 1 : 0);
		}
	M = M.canon();
	if(M.lead(_nrows) > _nrows)
	{
		cerr << "Attempting to invert a singular matrix!\n";
		return zero;
	}
	Matrix Ret(_nrows,_nrows);
	for(i=1; i<=_nrows; i++)
		for(j=1; j<=_nrows; j++)
			Ret(i,j) = M(i,j+_nrows);
	return Ret;
}

// Dummy definitions - for invokation of the templates for basic matrices
Matrix<int> _dr_matrix_m1(1,1);
Matrix<float> _dr_matrix_m2(1,1);
Vector<int> _dr_matrix_v1(1);
Vector<float> _dr_matrix_v2(1);
